--
-- VHDL course - Practical Phase III - Autumn 2002
--
-- Group 3: 
--  Jouni Osmala <josmala@cc.hut.fi>
--  Ilpo Jrvinen <ilpo.jarvinen@helsinki.fi>
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_signed.all;
use work.wnm_types.all;
use work.random.all;

-- Naming that guarantees first place in the entity list (easier GUI access
-- into it).
entity a_dsp_testbench is
end a_dsp_testbench;

architecture behave of a_dsp_testbench is
  -- 100 MHz clock...
  constant halfclk : time := 1 sec / (100 * 10 ** 6) / 2;
  
component wnm
   port(
      clk	  : in std_logic;
      reset : in std_logic;
      
      ref_time : in std_logic_vector(7 downto 0);
      ant_no   : in std_logic_vector(3 downto 0);
      sample_i : in std_logic_vector(5 downto 0);
      sample_q : in std_logic_vector(5 downto 0);
      
      block_select : in std_logic;
      write_enable : in std_logic;
      address      : in std_logic_vector(7 downto 0);
      data_in      : in std_logic_vector(11 downto 0);      
      data_out     : out std_logic_vector(11 downto 0);
      alarm        : out std_logic );
end component;

component DSP
    port (
	clk: in std_logic;
	reset: in std_logic;
	
	wnm_block_select: out std_logic;
	wnm_write_enable: out std_logic;
	wnm_address:out std_logic_vector(7 downto 0);
	wnm_data_out:out std_logic_vector(11 downto 0);
	wnm_data_in: in std_logic_vector(11 downto 0);
	wnm_alarm: in std_logic;

	command: in instruction );
end component;

-- WNM communication signals.
signal clk	  : std_logic := '0';
signal reset : std_logic := '0';
signal ref_time : std_logic_vector(7 downto 0) := x"00";
signal ant_no   : std_logic_vector(3 downto 0) := x"0";
signal sample_i : std_logic_vector(5 downto 0) := b"000000";
signal sample_q : std_logic_vector(5 downto 0) := b"000000";
signal block_select : std_logic := '0';
signal write_enable :  std_logic := '0';
signal address : std_logic_vector(7 downto 0) := x"00";
signal data_in : std_logic_vector(11 downto 0);  
signal data_out : std_logic_vector(11 downto 0);
signal alarm : std_logic;


-- Test types and constants...
constant verbose : integer := 1;

constant ant_no_max : integer := 16;
constant ref_time_max : integer := 128;

subtype i_index is integer range -32 to 31;
subtype q_index is integer range -32 to 31;


--
-- Resolved integer type r_integer, the resolution function returns the last non-zero value.
--
type integer_array is array (integer range <>) of integer;
function resolve_integer ( values : in integer_array ) return integer is
	variable result : integer := 0;
begin
	for i in values'range loop
		if values(i) /= 0 then
			result := values(i);
		end if;
	end loop;
	return result;
end function;
subtype r_integer is resolve_integer integer;

-- Noise total holder array (one slot for each antenna).
type    result_regs is array (reg_index) of r_integer;
constant clear_results : result_regs := (others => 0);

-- Signals for result testing process.
signal results : result_regs := clear_results; -- Signal has two sources (resolved).
signal validation : r_integer := 0;            -- Signal has two sources (resolved).
signal validation_ack : integer := 0;


-- Basic/Random test selector...
signal basic_tests_done : std_logic := '0';


-- DSP part
type dsp_ctrl_state is (NOP, TEST_ADDRS, EN_ALARM, READ_REGS, DIS_ALARM);
signal ctrl_state : dsp_ctrl_state;
signal dsp_command : instruction;


begin

  -- Reset signal is generated here.
  init: process is
  begin
    srand(1);                   -- random seed.
    wait for halfclk / 2;
    reset <= '1';
    wait for halfclk * 5;
    wait until clk = '0';
    wait until clk = '1';
    reset <= '0';
    while true loop
      wait until clk'event;
    end loop;
  end process init;

	
  -- Clock generator.
  clock : process is
  begin
    while true loop
      wait for halfclk;
      clk <= not(clk);
    end loop;
  end process clock;

  -- Ref_time and ant_no generation.
  ref_time_ant_no_gen: process is
    variable ref_ant: integer := 0;   -- MSBs: ref_time, LSBs: ant_no
  begin
    -- Drive initial values before the reset goes low.
    ant_no <= conv_std_logic_vector(ref_ant mod 16,4);
    ref_time <= conv_std_logic_vector(ref_ant/16,8);
    wait until reset = '0';
    
    while true loop
      if ref_ant < ant_no_max * ref_time_max - 1 then
        ref_ant := ref_ant + 1;
      else
   	ref_ant := 0;  -- Restart counters (wrap-around).
      end if;
      -- Wait until the next cycle begins.
      wait until clk = '1';
      ant_no <= conv_std_logic_vector(ref_ant mod 16,4);
      ref_time <= conv_std_logic_vector(ref_ant/16,8);
    end loop;
  end process ref_time_ant_no_gen;


  -- Basic tester
  --
  -- We generate constant i and q values using i_index and 
  -- q_index as range defination...
  test : process is
    variable i : integer := i_index'left;
    variable q : integer := q_index'left;
    variable total : integer := 0;
  begin
    -- Don't drive when random test is active.
    sample_i <= (others => 'Z');
    sample_q <= (others => 'Z');
    
    validation <= 0;
    results <= clear_results;
    if basic_tests_done = '1' then
      wait until basic_tests_done = '0';
    end if;
    
    while true loop
      sample_i <= conv_std_logic_vector(i,6);
      sample_q <= conv_std_logic_vector(q,6);
      wait until clk'event;
      validation <= 0;
      wait until ref_time'event or ant_no'event;

      if ref_time = x"00" and ant_no = x"0" then
        -- Calculate the results: (i^2 + q^2) * 128 / 128 = i^2 + q^2
   	total := i ** 2 + q ** 2;
   	for n in reg_index loop
          results(n) <= total;   -- Each register should have the same value
        end loop;
        
        validation <= 1;
        
        if verbose > 0 then
          report "Testing for" &
                 " i=" & i_index'image(i) & 
                 " q=" & q_index'image(q) &
                 " Sigma/n=" & integer'image(total)
            severity note;
        end if;

        -- Increment q and i.        
        if i /= i_index'right then
          i := i_index'succ(i);
        else
          i := i_index'left;
          if q /= q_index'right then
            q := q_index'succ(q);
          else

        	   -- This test is complete!
            report "Basic tests are complete!" 
              severity note;
            basic_tests_done <= '1';
            sample_i <= (others => 'Z');
	    sample_q <= (others => 'Z');
            wait until clk'event;
            validation <= 0;
            wait until validation_ack'event and validation_ack = 1;
            exit when true;
            
          end if;
        end if;
      end if;
			
    end loop;
  end process test;


  --
  -- Advanced (random) tester.  
  --
  rand_test: process is
    variable i : std_logic_vector(5 downto 0);
    variable q : std_logic_vector(5 downto 0);
    variable total : result_regs := clear_results;
    variable idx : integer;
  begin
    -- Don't drive when basic test is active.
    sample_i <= (others => 'Z');
    sample_q <= (others => 'Z');

    if basic_tests_done = '0' then
      wait until basic_tests_done = '1';
    end if;
    
    while true loop
      -- Get i & q and calculate sums.
      i := conv_std_logic_vector(rand,6);
      q := conv_std_logic_vector(rand,6);
      sample_i <= i;
      sample_q <= q;
      idx := conv_integer(unsigned(ant_no));
      if idx <= reg_index'high then
        total(idx) := total(idx) + conv_integer(signed(i)) ** 2 + 
        				   conv_integer(signed(q)) ** 2;
      end if;
      
      wait until clk'event;
      validation <= 0;
      wait until ref_time'event or ant_no'event;
      

      if ref_time = x"00" and ant_no = x"0" then
        for n in reg_index loop
          total(n) := total(n) / 128;
          results(n) <= total(n);
        end loop;
        
        if verbose > 0 then
           report "Testing for i,q=random"
             severity note;
        elsif verbose > 1 then
          for idx in reg_index loop
            report "Testing for regs " & integer'image(idx) &
                   " i=random" & 
                   " q=random" & 
                   " Sigma/n=" & integer'image(total(idx))
              severity note;
          end loop;
        end if;
        validation <= 1;
        sample_i <= (others => 'Z');
	sample_q <= (others => 'Z');
	total := clear_results;
        
      end if;
      
    end loop;
  end process rand_test;
	

dsp_ctrl: process is
  variable reg_idx : std_logic_vector(7 downto 0) := x"00";
  variable next_state : dsp_ctrl_state;
  variable rnd_num : std_logic_vector(2 downto 0);
begin
  ctrl_state <= NOP;
  dsp_command.opcode <= dsp_nop;
  dsp_command.reg1 <= x"00";
  dsp_command.reg2 <= x"00";
  dsp_command.reg3 <= x"00";
  wait until reset = '0';
  while true loop
    case ctrl_state is
      when NOP =>
        dsp_command.opcode <= dsp_nop;
        ctrl_state <= READ_REGS;
        reg_idx := x"00";
      when TEST_ADDRS =>
        dsp_command.opcode <= wnm_read;
        dsp_command.reg1 <= reg_idx;
        dsp_command.reg3 <= x"00";
        if reg_idx = x"ff" then
          ctrl_state <= EN_ALARM;
          reg_idx := x"00";
        else
          reg_idx := reg_idx + 1;
        end if;
      when EN_ALARM =>
        dsp_command.opcode <= wnm_write_im;
        dsp_command.reg1 <= x"01";
        dsp_command.reg2 <= x"01";
        ctrl_state <= READ_REGS;
        reg_idx := x"00";
      when DIS_ALARM =>
        dsp_command.opcode <= wnm_write_im;
        dsp_command.reg1 <= x"01";
        dsp_command.reg2 <= x"00";
        ctrl_state <= READ_REGS;
        reg_idx := x"00";
      when READ_REGS =>
        if reg_idx = 0 then
          reg_idx := x"f1";  -- Address of the first register.
        end if;
        dsp_command.opcode <= wnm_read;
        dsp_command.reg1 <= reg_idx;
        dsp_command.reg3 <= x"00";
        if reg_idx = x"fc" then
          -- Next state is randomly generated. AFAIK fulfils the specs. :-)
          rnd_num := conv_std_logic_vector(rand,3);
          case rnd_num is
            when b"000" => next_state := TEST_ADDRS;
            when b"001" => next_state := EN_ALARM;
            when b"010" => next_state := DIS_ALARM;
            when b"011" => next_state := READ_REGS;
            when others => next_state := NOP;
          end case;
          ctrl_state <= next_state;
          reg_idx := x"00";
        else
          reg_idx := reg_idx + 1;
        end if;
    end case;
    wait until rising_edge(clk);
  end loop;
end process dsp_ctrl;
		
			
w: component wnm
  port map(
     clk => clk,
     reset => reset,
     ref_time => ref_time,
     ant_no => ant_no,
     sample_i =>sample_i,
     sample_q =>sample_q,
     block_select => block_select,
     write_enable => write_enable,
     address => address,
     data_in => data_in,
     data_out => data_out,
     alarm => alarm);

d: component dsp
  port map(
     clk => clk,
     reset => reset,
     wnm_block_select => block_select,
     wnm_write_enable => write_enable,
     wnm_address => address,
     wnm_data_in => data_out,
     wnm_data_out => data_in,
     wnm_alarm => alarm,
     command => dsp_command);
     
end behave;
