--
-- VHDL course - Practical Phase III - Autumn 2002
--
-- Group 3: 
--  Jouni Osmala <josmala@cc.hut.fi>
--  Ilpo Jrvinen <ilpo.jarvinen@helsinki.fi>
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use work.wnm_types.all;


entity data_access is
  port (
    clk          : in  std_logic;
    reset        : in  std_logic;

    block_select : in  std_logic;
    write_enable : in  std_logic;
    address      : in  std_logic_vector(7 downto 0);
    data_in      : in  std_logic_vector(11 downto 0);
    
    data_out     : out std_logic_vector(11 downto 0);
    alarm        : out std_logic;

    store_result : in  std_logic;
    reg_in       : in  result_array);
end data_access;


architecture behave of data_access is

  component mux_2_to_1
    generic (
      bits : integer);
    port (
      d0 : in  std_logic_vector(bits-1 downto 0);   -- Data input 0
      d1 : in  std_logic_vector(bits-1 downto 0);   -- Data input 1
      s  : in  std_logic;                           -- Select
      q  : out std_logic_vector(bits-1 downto 0));  -- Data out
  end component;
  
  -- Alarm status reg.
  signal alarm_in        : std_logic;
  signal alarm_status    : std_logic;

  -- Results after division.
  signal result    : result_array;
  signal result_in : result_array;

begin  -- behave

  -----------------------------------------------------------------------------
  -- Clock edge and reset detection.
  --
  edge: process (reset, clk) is
    variable idx : integer range 0 to 15;
  begin
  
    if reset = '1' then
      data_out <= (others => '0');
      alarm <= '0';
      alarm_status <= '0';
      for i in reg_index loop
        result(i) <= result_zero;
      end loop;  -- i
      
    elsif rising_edge(clk) then
      alarm <= alarm_in;
      result <= result_in;

      -- data_in and data_out:
      data_out <= (others => '0');
      if block_select = '1' then
        if address = x"01" then
          -- Check that command is valid, only x"000" or x"001" are 
          -- acceptable...
          if write_enable = '1' then
            if data_in(11 downto 1) = b"00000000000" then
              alarm_status <= data_in(0);
            end if;
          else
            data_out <= (0 => alarm_status, others => '0');
          end if;

        elsif address(7 downto 4) = x"f" and 
              not (address(3 downto 0) = x"0")then   -- Antenna selected (f1-fc(-ff))?
          idx := conv_integer(unsigned(address(3 downto 0))) - 1;
          if 0 <= idx and idx < reg_count then
            if write_enable = '0' then
              data_out <= result(idx);
            else
              -- Should result be read-only?
              result(idx) <= data_in;
            end if;
          end if;
        end if;
          
      end if;                           -- b_s = '1'
    end if;                             -- edge(clk)
    
  end process edge;


  -----------------------------------------------------------------------------
  -- The alarm is controlled in here.
  -- When edge comes, we have already pre-calculated the value of the
  -- alarm for the next cycle into alarm_in signal.
  --
  alarm_update: process (alarm_status, reset, result_in) is
  begin
    alarm_in <= '0';
    if reset = '1' then
      alarm_in <= '0';
        
    -- Check if alarm is active (alarm control register is 1).
    elsif alarm_status = '1' then
      alarm_in <= '0';
      for i in reg_index loop
        if conv_integer(unsigned(result_in(i))) > 1023 then
          alarm_in <= '1';
        end if;
      end loop;
    else
      alarm_in <= '0';
    end if;
  end process alarm_update;


  -- This loop generates one MUX for each result register which
  -- is updated on the cycle with high store_result.
  resultgen: for i in reg_index generate
    begin
      mux: component mux_2_to_1
        generic map (
          bits => result_size)
        port map (
          d0 => result(i),
          d1 => reg_in(i),
          s  => store_result,
          q  => result_in(i));
  end generate resultgen;

end behave;
