--
-- VHDL course - Practical Phase III - Autumn 2002
--
-- Group 3: 
--  Jouni Osmala <josmala@cc.hut.fi>
--  Ilpo Jrvinen <ilpo.jarvinen@helsinki.fi>
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use work.wnm_types.all;

entity wnm is
  port(
    clk   : in std_logic;
    reset : in std_logic;

    ref_time : in std_logic_vector(7 downto 0);
    ant_no   : in std_logic_vector(3 downto 0);
    sample_i : in std_logic_vector(5 downto 0);
    sample_q : in std_logic_vector(5 downto 0);

    block_select : in std_logic;
    write_enable : in std_logic;
    address      : in std_logic_vector(7 downto 0);
    data_in      : in std_logic_vector(11 downto 0);

    data_out : out std_logic_vector(11 downto 0);
    alarm    : out std_logic);
end entity wnm;


architecture behave of wnm is

  -----------------------------------------------------------------------------
  -- Components.
  -----------------------------------------------------------------------------
  component data_access
    port (
      clk   : in std_logic;
      reset : in std_logic;

      block_select : in std_logic;
      write_enable : in std_logic;
      address      : in std_logic_vector(7 downto 0);
      data_in      : in std_logic_vector(11 downto 0);

      data_out : out std_logic_vector(11 downto 0);
      alarm    : out std_logic;

      store_result : in std_logic;
      reg_in       : in result_array);
  end component;

  component multadd
    generic (
      input_size : integer := 6;
      output_size : integer := 19);
    port (
      a : in std_logic_vector(5 downto 0);
      b : in std_logic_vector(5 downto 0);
      c : in reg_t;
      r : out reg_t);
  end component;


  component decoder
    generic (
      inputs : integer;                                -- Width of the input
      outputs : integer);                              -- Width of the output
    port (
      din  : in std_logic_vector(inputs-1 downto 0);   -- Data input
      dout : out std_logic_vector(outputs-1 downto 0); -- Data output
      err  : out std_logic);                           -- din overflow.
  end component;
  
  component tristate_buffer
    generic (
      bits : integer := 1);                          -- Width of the data
    port (
      ena  : in  std_logic;                          -- dout driven?
      din  : in  std_logic_vector(bits-1 downto 0);  -- data input
      dout : out std_logic_vector(bits-1 downto 0)); -- data output to bus
  end component;

  component regn
    generic (
      bits : integer := 8);                              -- Width of the reg

    port (
      clk   : in  std_logic;
      reset : in  std_logic;                             -- Asynchronious reset
      load  : in  std_logic;                             -- Load din into reg
      clear : in  std_logic;                             -- Synchronious clear

      din   : in     std_logic_vector(bits-1 downto 0);  -- Data input
      dout  : buffer std_logic_vector(bits-1 downto 0)); -- Data output & store
  end component;


  -----------------------------------------------------------------------------
  -- Internal signals.
  -----------------------------------------------------------------------------
    
  signal result_out : result_array;                -- Result interface.
  signal bus_from_reg, bus_to_reg : reg_t;         -- Integrator datapath.
  signal ant_no_dec : std_logic_vector(reg_index); -- Decoded antenna number.
  signal update : std_logic;                       -- Result loading control.
  signal no_antenna : std_logic;                   -- Currently unused.



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
begin
  
  ---------------------------------------------------------------------------
  -- When both ref_time and ant_no wrap, we load the new
  -- results and clear the registers.
  update_results: process (ref_time, ant_no)
  begin
    if ref_time = x"7f" and ant_no = x"f" then
      update <= '1';
    else
      update <= '0';
    end if;
  end process update_results;

  -- ant_no is decoded into individidual signals.
  antdec: component decoder
    generic map (
      inputs => 4,
      outputs => 12)
    port map (
      din => ant_no,
      dout => ant_no_dec,
      err => no_antenna);

  
  -- Data store and access module.
  da: component data_access
    port map (
      clk          => clk,
      reset        => reset,
      block_select => block_select,
      write_enable => write_enable,
      address      => address,
      data_in      => data_in,
      data_out     => data_out,
      alarm        => alarm,
      store_result => update,
      reg_in       => result_out);

  -- Regs-to-Integrator: Tri-state buffers are used to form a bus structure.
  -- Integrator-to-regs: The result is wired to each register but only one has
  --                     high value in load port.
  -- Regs-to-Result-Div: The divide "operation" is a simple wiring problem,
  --                     the LSB wires are "dropped" from the signal.
  --                     We can safely wire the registers into results through
  --                     the divide as the last antenna is not a real one.
  busgen: for i in reg_index generate
    signal temp : reg_t;
  begin
    -- Reg & i2r connections.
    reg: component regn
      generic map (
        bits => reg_size)
      port map (
        clk   => clk,
        reset => reset,
        load  => ant_no_dec(i),
        clear => update,
        din   => bus_to_reg,
        dout  => temp);

    -- r2i
    tri: component tristate_buffer
      generic map (
        bits => reg_size)
      port map (
        ena => ant_no_dec(i),
        din => temp,
        dout => bus_from_reg);
    
    -- div
    result_out(i) <= temp(reg_size-1 downto reg_size-result_size);

  end generate busgen;

  -- The integrator
  --     
  -- THINKME: We drive 'Z' to the c when antenna is not valid, is this a good
  -- way to do things or should we use no_antenna signal to drive '0' instead
  -- of the 'Z'...
  ma: component multadd
    generic map (
      input_size => 6,
      output_size => reg_size)
    port map (
      a => sample_i,
      b => sample_q,
      c => bus_from_reg,
      r => bus_to_reg);
  
end behave;
